// -*- C++ -*-
/*!
 * @file  rtc_factorization.h
 * @brief Image processing of the factorization
 * @date  $Date$
 *
 * @author EAST JAPAN ACCOUNTING CENTER CO.,LTD.
 *
 * BSD 3-clause "New" or "Revised" License
 *
 * $Id$
 */

#ifndef RTC_FACTORIZATION_H
#define RTC_FACTORIZATION_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

// add to code
#include <iostream>
#include <string>
#include <stdio.h>
#include "windows.h"

// add to code for OpenCV
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

// add to code
#include "local_basic_type.h"
#include "2DCDP/2DCDP4.h"
#include "2DCDP/2DCDP4FTV.h"
#include "FactorizationController/factorizationcontroller.h"
#include "CDPObjectFormatter/CDPObjectFormatter.h"
#include "Object/ObjectMesh.h"

// Service implementation headers
// <rtc-template block="service_impl_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">
#include "Rlt2dcdpIFStub.h"
#include "RltObjectIFStub.h"

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="port_stub_h">
using namespace Rlt2dcdpIF;
using namespace RltObjectIF;
// </rtc-template>

using namespace RTC;

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

// add to code
/*=================================================================================================
  Macro definition
 ================================================================================================*/
#define _DEBUG_SW_

//#define RESULT_OK                    (1)
//#define RESULT_NG                    (0)

#define FACTOR_STAGE_INIT              (0)
#define FACTOR_STAGE_MIN               (FACTOR_STAGE_INIT)
#define FACTOR_STAGE1                  (1)
#define FACTOR_STAGE2                  (2)
#define FACTOR_STAGE3                  (3)
#define ERROR_DETECTED                 FACTOR_STAGE3
#define FACTOR_STAGE_MAX               (FACTOR_STAGE3+1)

#define IMAGES_LENGTH_2                (2)
#define FORMAT_COLOR24                 (24)
#define CHANNEL_NUM                    (3)

#define XIMG_120                       (120)
#define YIMG_90                        (90)

#define XIMG_320                       (320)
#define YIMG_240                       (240)

#define _ON_INITIALIZE                 ("onInitialize execution")
#define _ON_ACTIVATED                  ("onActivated execution")
#define _ON_DEACTIVATED                ("onDeactivated execution")
#define _ON_ABORTING                   ("onAborting execution")
#define _ON_RESET                      ("onReset execution")

#define FACTORIZATION_SUCCESS          ("Factorization RTC process was successful")

#define _ERROR_IMAGE_COLOR             ("ERROR: This image data is not implemented in this program")
#define _ERROR_IMAGE_EMPTY             ("ERROR: Empty error of image data")
#define _ERROR_IMAGE_SIZE              ("ERROR: This image data is outside the size target")
#define _ERROR_IMAGE_DATA              ("ERROR: Image data error")
#define _ERROR_ALLOCATION              ("ERROR: Memory allocation error")
#define _STAGE_ERROR                   ("ERROR: Wrong value of the stage variable")
#define _ERROR_OBJECT_CREATION         ("ERROR: Can't create Object File")
#define _ERROR_OBJ_CREATE              ("ERROR: Can't create object")
#define _ERROR_HEAP_MEM_MCORSP         ("ERROR: Can't heap Memory m_corsp")
#define _ERROR_HEAP_MEM_PICKBUF        ("ERROR: Can't heap Memory pickbuf")
#define _ERROR_HEAP_MEM_PICKUP1        ("ERROR: Can't heap Memory pickup1")
#define _ERROR_HEAP_MEM_PICKUP2        ("ERROR: Can't heap Memory pickup2")
#define _ERROR_WRITE_OBJ               ("ERROR: ResultObject data port is not connected. please check the connect of data port!")
#define _ERROR_WRITE_IMG               ("ERROR: Image data port is not connected. please check the connect of data port!")
#define _ERROR_RESET                   ("ERROR: Error was detected. Please reset the factorization RTC by the RT System Editor.")

#define KIND_IMAGE_INIT                (0)
#define KIND_IMAGE_DATA1               (1)
#define KIND_IMAGE_DATA2               (2)

#define RESULT_2DCDP_1                 (0)
#define RESULT_2DCDP_2                 (1)

#define IMG_POS_C                      (0)
#define IMG_POS_L                      (1)
#define IMG_POS_R                      IMG_POS_L
#define IMG_INP_C                      (1)

#define KIND_POS_L                     (0)
#define KIND_POS_R                     (1)

#define SEQ_NO_1                       (0)
#define SEQ_NO_2                       (1)

#define ERR_CODE_INIT                  (0)
#define ERR_CODE_MIN                   ERR_CODE_INIT
#define ERR_IMAGE_DATA                 (1)
#define ERR_ALLOCATE_M                 (2)
#define ERR_OBJ_CREATE                 (3)
#define ERR_CODE_MAX                   (ERR_OBJ_CREATE+1)

enum{ NUM_IMAGE = 3, };
enum IMAGE_TYPE{ LEFT =0, CENTER, RIGHT, };

/*!
 * @class rtc_factorization
 * @brief Image processing of the factorization
 *
 * Image processing of the factorization
 *
 * multi-cameraimages
 *
 * -
 *
 * http://opencv.org/
 *
 */
class rtc_factorization
  : public RTC::DataFlowComponentBase
{
 public:
  /*!
   * @brief constructor
   * @param manager Maneger Object
   */
  rtc_factorization(RTC::Manager* manager);

  /*!
   * @brief destructor
   */
  ~rtc_factorization();

  // <rtc-template block="public_attribute">
  
  // </rtc-template>

  // <rtc-template block="public_operation">
  
  // </rtc-template>

  /***
   * Initialization of the factorization RTC
   *
   * The initialize action (on CREATED->ALIVE transition)
   * formaer rtc_init_entry() 
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onInitialize();

  /***
   * 
   *
   * The finalize action (on ALIVE->END transition)
   * formaer rtc_exiting_entry()
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onFinalize();

  /***
   *
   * The startup action when ExecutionContext startup
   * former rtc_starting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  /***
   *
   * The shutdown action when ExecutionContext stop
   * former rtc_stopping_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  /***
   * Activate the factorization RTC
   *
   * The activated action (Active state entry action)
   * former rtc_active_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  /***
   * Deactivate the factorization RTC
   *
   * The deactivated action (Active state exit action)
   * former rtc_active_exit()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  /***
   * Periodic task of the factorization RTC
   *
   * The execution action that is invoked periodically
   * former rtc_active_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  /***
   * onAborting the factorization RTC
   *
   * The aborting action when main logic error occurred.
   * former rtc_aborting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  /***
   *
   * The error action in ERROR state
   * former rtc_error_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  /***
   * Reset initialization of the factorization RTC
   *
   * The reset action that is invoked resetting
   * This is same but different the former rtc_init_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);
  
  /***
   *
   * The state update action that is invoked after onExecute() action
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  /***
   *
   * The action that is invoked when execution context's rate is changed
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // <rtc-template block="protected_attribute">
  
  // </rtc-template>

  // <rtc-template block="protected_operation">
  
  // </rtc-template>

  // Configuration variable declaration
  // <rtc-template block="config_declare">

  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::MultiCameraImages m_in_img1;
  /*!
   * This port inputs for the factorization image data.
   * - Type: RTC::MultiCameraImages
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<RTC::MultiCameraImages> m_in_img1In;
  RTC::MultiCameraImages m_in_img2;
  /*!
   * This port inputs for the factorization image data.
   * - Type: RTC::MultiCameraImages
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<RTC::MultiCameraImages> m_in_img2In;
  Rlt2dcdpIF::Result2dcdp m_rlt_2dcdp;
  /*!
   * This port inputs the result data of 2DCDP.
   * - Type: Rlt2dcdpIF::Result2dcdp
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<Rlt2dcdpIF::Result2dcdp> m_rlt_2dcdpIn;
  
  // </rtc-template>


  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  RTC::MultiCameraImages m_out_img;
  /*!
   * This port outputs for the factorization image data.
   * - Type: RTC::MultiCameraImages
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  OutPort<RTC::MultiCameraImages> m_out_imgOut;
  RltObjectIF::ResultObject m_rlt_obj;
  /*!
   * This port outputs the result data of the factorization.
   * - Type: RltObjectIF::ResultObject
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  OutPort<RltObjectIF::ResultObject> m_rlt_objOut;
  
  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">
  
  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">
  
  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">
  
  // </rtc-template>

 private:
  // <rtc-template block="private_attribute">
  
  // </rtc-template>

  // <rtc-template block="private_operation">
  
  // </rtc-template>

  // add on prototype of function
  void nvdFactorOnInitialize( void );
  void nvdFactorOnActivated( void );
  void nvdFactorOnDeactivated( void );
  void nvdFactorOnExecute( void );
  void nvdFactorOnAborting( void );
  void nvdFactorOnReset( void );
  void nvdOutCameraImagesCreation( void );
  void nvdRelease( void );

  void nvdModeMatrixExecute( uhword_type, uhword_type );
  void nvdSaveImageData( RTC::MultiCameraImages * );
  void nvdCopy_pickup1( uword_type );
  void nvdCopy_pickup2( uword_type );
  void nvdSetStageControl( ubyte_type );
  void nvdResultObjectCreation( OBJ_MESH::Mesh &mesh );
  void nvdResultObjectVertex( OBJ_MESH::Mesh &mesh );
  void nvdResultObjectTexture( OBJ_MESH::Mesh &mesh );
  void nvdResultObjectFace( OBJ_MESH::Mesh &mesh );
  void nvdResultObjectType( OBJ_MESH::Mesh &mesh );
  void nvdSetErrorCode( uhword_type );
  void nvdOutputErrorMessages( void );

  uword_type uwMainMatrixExecute( void );
  uword_type uwSetCameraImages( cv::Mat, uword_type );

  uhword_type uhwInEvCheckofMatrix( void );
  uhword_type uhwProc_InputImage( void );
  uhword_type uhwProc_InputResult1( void );
  uhword_type uhwProc_InputResult2( void );
  uhword_type uhwCheckofResult2dcdp( void );
  uhword_type uhwCheckofInputImage( void );

  bool_type bCheckMultiCameraImages( RTC::MultiCameraImages );
  bool_type bCheckMultiCameraImagesOfSize( uword_type, uword_type );
  bool_type bAllocateMemory_m_corsp( uword_type );
  bool_type bAllocateMemory_pickbuf( uword_type );
  bool_type bAllocateMemory_pickup1( uword_type );
  bool_type bAllocateMemory_pickup2( uword_type );
  bool_type bCreateMesh( OBJ_MESH::Mesh &mesh );
  bool_type bSetTrajectData1( TrajectData& data, const int& index );
  bool_type bSetTrajectData2( TrajectData& data, const int& index );
  bool_type bSetTrajectDataCenter( TrajectData& data, const int& index );

	// variables
  struct rtcPos2D{
    bool_type  isData;
    uword_type i;
    uword_type j;
    inline void Init(){ isData = false; i = 0; j = 0; }
  };

  struct rtcPICK
  {
    shword_type x;
    shword_type y;
    ubyte_type  x0;
    ubyte_type  y0;
  };

  rtcPos2D*  rtc_m_corsp;
  char_type* pickbuf;
  rtcPICK*   pickup1;
  rtcPICK*   pickup2;

	// variables of matrix
	uhword_type m_uhwInEvent;
	uhword_type m_uhwMatrixMode;
	uhword_type m_uhwImageData;
	uhword_type m_uhwErrorCode;

  IplImage m_IplImage;

  ubyte_type m_ubStageCtrl;
  ubyte_type m_ubStageCtrlBackUp;

  udword_type inpH;
  udword_type inpW;
  udword_type refH;
  udword_type refW;

  uword_type m_pix_num;

  OBJ_MESH::Mesh g_obj_mesh;
  CDP::IMG inp;
  CDP::IMG ref;
  CDP::Color bgColor;

};


extern "C"
{
  DLL_EXPORT void rtc_factorizationInit(RTC::Manager* manager);
};

#endif // RTC_FACTORIZATION_H
